var responseBuffer = [];
var contentLength = 0;
var NGINX_PLUS_CONTEXT = "nginxplus";
var numSubrequests;
var keyVal = null;

var ResponseMetrics = [
    "1xx",
    "2xx",
    "3xx",
    "4xx",
    "5xx"
]

var SessionMetrics = [
    "2xx",
    "4xx",
    "5xx"
]

var HealthCheckMetrics = {
    "checks": {
        "help": "Total health check requests",
        "type": "counter"
    },
    "fails": {
        "help": "Failed health checks",
        "type": "counter"
    },
    "unhealthy": {
        "help": "How many times the server became unhealthy (state 'unhealthy')",
        "type": "counter"
    },
}

var UpstreamServerMetrics = {

    "active": {
        "help": "Active client connections",
        "type": "gauge"
    },

    "fails": {
        "help": "Number of unsuccessful attempts to communicate with the server",
        "type": "counter"
    },

    "header_time": {
        "help": "Average time to get the response header from the server",
        "type": "gauge"
    },

    "health_checks": {
        "handler": prepend_metric_handler,
        "metrics": HealthCheckMetrics
    },
    "received": {
        "help": "Bytes received from clients",
        "type": "counter"
    },
    "requests": {
        "help": "Total client requests",
        "type": "counter"
    },
    "response_time": {
        "help": "Average time to get the full response from the server",
        "type": "gauge"
    },
    "responses": {
        "help": "The total number of responses obtained from this server",
        "type": "counter",
        "handler": labeled_metric_handler,
        "metrics": ResponseMetrics
    },
    "sent": {
        "help": "Bytes sent to clients",
        "type": "counter"
    },
    "state": {
        "help": "Current state",
        "type": "gauge",
        "printFilter": convert_state
    },
    "unavail": {
        "help": "How many times the server became unavailable for client requests (state 'unavail') due to the number of unsuccessful attempts reaching the max_fails threshold",
        "type": "counter"
    }
};

var UpstreamMetrics = {

    "keepalive": {
        "help": "Idle keepalive connections",
        "type": "gauge"
    },

    "peers": {
        "handler": upstream_server_handler,
        "metrics": UpstreamServerMetrics,
        "context": "upstream_server",
        "specifierContext": "server"
    },

    "zombies": {
        "help": "Servers removed from the group but still processing active client requests",
        "type": "gauge"
    },
};

var ServerZoneMetrics = {
    "discarded": {
        "help": "Requests completed without sending a response",
        "type": "counter"
    },
    "processing": {
        "help": "Client requests that are currently being processed",
        "type": "gauge"
    },
    "received": {
        "help": "Bytes received from clients",
        "type": "counter"
    },
    "requests": {
        "help": "Total client requests",
        "type": "counter"
    },
    "responses": {
        "help": "Total responses sent to clients",
        "type": "counter",
        "handler": labeled_metric_handler,
        "metrics": ResponseMetrics
    },
};

var StreamUpstreamServerMetrics = {
    "active": {
        "help": "Active client connections",
        "type": "gauge"
    },
    "connect_time": {
        "help": "Average time to connect to the upstream server",
        "type": "gauge"
    },
    "connections": {
        "help": "Total number of client connections forwarded to this server",
        "type": "counter"
    },
    "fails": {
        "help": "Number of unsuccessful attempts to communicate with the server",
        "type": "counter"
    },
    "first_byte_time": {
        "help": "Average time to receive the first byte of data",
        "type": "gauge"
    },
    "health_checks": {
        "handler": prepend_metric_handler,
        "metrics": HealthCheckMetrics
    },
    "received": {
        "help": "Bytes received from clients",
        "type": "counter"
    },
    "response_time": {
        "help": "Average time to receive the last byte of data",
        "type": "gauge"
    },
    "sent": {
        "help": "Bytes sent to clients",
        "type": "counter"
    },
    "state": {
        "help": "Current state",
        "type": "gauge",
        "printFilter": convert_state
    },
    "unavail": {
        "help": "How many times the server became unavailable for client requests (state 'unavail') due to the number of unsuccessful attempts reaching the max_fails threshold",
        "type": "counter"
    },

}

var StreamUpstreamMetrics = {
    "zombies": {
        "help": "Servers removed from the group but still processing active client requests",
        "type": "gauge"
    },
    "peers": {
        "handler": upstream_server_handler,
        "metrics": StreamUpstreamServerMetrics,
        "context": "stream_upstream_server",
        "specifierContext": "server"
    }
};

var StreamServerZoneMetrics = {
    "connections": {
        "help": "Total connections",
        "type": "counter"
    },
    "discarded": {
        "help": "Connections completed without creating a session",
        "type": "counter"
    },
    "processing": {
        "help": "Client connections that are currently being processed",
        "type": "gauge"
    },
    "received": {
        "help": "Bytes received from clients",
        "type": "counter"
    },
    "sent": {
        "help": "Bytes sent to clients",
        "type": "counter"
    },
    "sessions": {
        "help": "Total sessions completed",
        "type": "counter",
        "handler": labeled_metric_handler,
        "metrics": SessionMetrics
    }
};

var ConnectionMetrics = {
    "accepted": {
        "help": "Accepted client connections",
        "type": "counter"
    },
    "active": {
        "help": "Active client connections",
        "type": "gauge"
    },
    "dropped": {
        "help": "Dropped client connections",
        "type": "counter"
    },
    "idle": {
        "help": "Idle client connections",
        "type": "gauge"
    }
};

var HttpRequestMetrics = {
    "total": {
        "help": "Total http requests",
        "type": "counter"
    },
    "current": {
        "help": "Current http requests",
        "type": "gauge"
    }
};

var SslMetrics = {
    "handshakes": {
        "help": "Successful SSL handshakes",
        "type": "counter"
    },
    "handshakes_failed": {
        "help": "Failed SSL handshakes",
        "type": "counter"
    },
    "session_reuses": {
        "help": "Session reuses during SSL handshake",
        "type": "counter"
    },
}


var ConnectionContext = {
    "handler": default_metric_callback,
    "metrics": ConnectionMetrics,
    "context": "connections"
}

var SslContext = {
    "handler": default_metric_callback,
    "metrics": SslMetrics,
    "context": "ssl"
}

var HttpRequestContext = {
    "handler": default_metric_callback,
    "metrics": HttpRequestMetrics,
    "context": "http_requests"
}

var UpstreamContext = {
    "handler": named_metric_callback,
    "metrics": UpstreamMetrics,
    "context": "upstream"
}

var ServerZoneContext = {
    "handler": named_metric_callback,
    "metrics": ServerZoneMetrics,
    "context": "server_zone"
}

var StreamUpstreamContext = {
    "handler": named_metric_callback,
    "metrics": StreamUpstreamMetrics,
    "context": "stream_upstream",
    "specifierContext": "upstream"
}

var StreamServerZoneContext = {
    "handler": named_metric_callback,
    "metrics": StreamServerZoneMetrics,
    "context": "stream_server_zone",
    "specifierContext": "server_zone"
}

var Metrics = [
    [
        "/api/3/connections",
        ConnectionContext
    ],
    [
        "/api/3/ssl",
        SslContext
    ],
    [
        "/api/3/http/requests",
        HttpRequestContext
    ],
    [
        "/api/3/http/upstreams",
        UpstreamContext,
        replace_with_keyval
    ],
    [
        "/api/3/http/server_zones",
        ServerZoneContext
    ],
    [
        "/api/3/stream/upstreams",
        StreamUpstreamContext,
        replace_with_keyval
    ],
    [
        "/api/3/stream/server_zones",
        StreamServerZoneContext
    ]
]

function metrics(r) {
    numSubrequests = Metrics.length
    Metrics.forEach(function(m) {
        ngx_subrequest.apply(null, [r].concat(m))
    })
}

// ngx_subrequest makes a subrequest to a designated url and then passes the
// resulting object to a handler to build the response for the metrics passed in
function ngx_subrequest(r, url, objContext, filterHandler) {
    r.subrequest(url, { method: 'GET' }, function (res) {
        if (res.status >= 300 || res.responseBody == "{}") {
            if (--numSubrequests == 0) {
                ngx_send_response(r);
            } else {
                return;
            }
        }
        contentLength += res.responseBody.length;

        var obj = JSON.parse(res.responseBody);
        if (filterHandler) {
            filterHandler(r, obj, objContext);
        } else {
            var buf = process_obj(obj, objContext);
            responseBuffer.push(buf);
            if (--numSubrequests == 0) {
                ngx_send_response(r);
            }
        }
    });
}

// replace_with_keyval makes a subrequest to the keyval specified by the variable
// "prom_keyval" and replaces all object names if they are in the keyval map
// keyval map should take the form replacement : object-to-replace
function replace_with_keyval(r, obj, objContext) {
    var keyValUpstream = r.variables.prom_keyval;
    if (!keyVal && keyValUpstream) {
        r.subrequest("/api/3/http/keyvals/" + keyValUpstream, { method: 'GET' },
            function (res) {
                if (!keyVal) {
                    if (res.status >= 300) {
                        keyVal = NaN;
                    } else {
                        keyVal = JSON.parse(res.responseBody);
                        var keyValEnum = Object.entries(keyVal);
                        for (var i = 0; i < keyValEnum.length; i++) {
                            var upstreamMap = keyValEnum[i];
                            if (obj[upstreamMap[1]] != null) {
                                obj[upstreamMap[0]] = obj[upstreamMap[1]];
                                delete obj[upstreamMap[1]];
                            }
                        }
                    }
                }
                var buf = process_obj(obj, objContext);
                responseBuffer.push(buf);
                if (--numSubrequests == 0) {
                    ngx_send_response(r);
                }
            });
    }
}

// process_obj is the callback from the ngx_subrequest function and is
// responsible for printing the metric help text as well as calling the handler
// associated with a certain metric
function process_obj(obj, objContext) {
    var buffer = [];
    var metricsEnum = Object.entries(objContext.metrics);
    var metricContext = objContext.context;
    var handler = objContext.handler;
    for (var i = 0; i < metricsEnum.length; i++) {
        var metric = metricsEnum[i];
        // Print metric information
        if (metric[1].help != null) {
            buffer.push(print_metric_help_text(metricContext, metric));
        }

        var specifierContext = objContext.context;
        if (objContext.specifierContext != null) {
            specifierContext = objContext.specifierContext;
        }
        // Call if metric has associated handler
        if (metric[1].handler != null) {
            buffer.push(metric[1].handler(obj, metric, metricContext,
                specifierContext));
        } else {
            // Else call default handler
            buffer.push(handler(obj, metric, metricContext, specifierContext));
        }
    };
    return buffer.join("");
}

// default_metric_callback assumes that there is only a single object of metrics
// to be printed. It returns metrics without a specifier attached to them
function default_metric_callback(obj, metric, context) {
    return build_metric(context, metric[0], obj[metric[0]], null);
}

// named_metric_callback assumes that there are multiple elements to be returned,
// each formatted as a name : [object] pair in which the name will be used
// in conjunction with the specifierContext to build a specifier;
function named_metric_callback(obj, metric, context, specifierContext) {
    var buffer = [];
    var arrayEnum = Object.entries(obj);

    for (var j = 0; j < arrayEnum.length; j++) {
        var arrayKey = arrayEnum[j][0]
        var arrayObj = arrayEnum[j][1];
        var metric_val = arrayObj[metric[0]];
        var specifierObj = {};
        if (metric[1].printFilter != null) {
            metric_val = metric[1].printFilter(arrayObj[metric[0]]);
        }
        // On iterations after initial, specifier object may be set, representing
        // past base contexts to be printed
        if (arrayObj.specifierObj != null) {
            specifierObj = arrayObj.specifierObj;
        }
        specifierObj[specifierContext] = arrayKey;
        buffer.push(build_metric(context, metric[0], metric_val, specifierObj))
        delete specifierObj[specifierContext];
    }
    return buffer.join("");
}

// upstream_server_handler is responsible for building the name : object pair
// that named_metric_callback expects, except with peer servers instead of
// upstreams
function upstream_server_handler(obj, metric, context, specifierContext) {
    var nextObj = {};
    var upstreamsEnum = Object.entries(obj);
    var metric_name = metric[0];
    var metric_obj = metric[1];

    for (var upstreamIdx = 0; upstreamIdx < upstreamsEnum.length; upstreamIdx++) {
        var upstream = upstreamsEnum[upstreamIdx][1];
        var specifierObj = {};
        // Add specifier for upstream to object to attach to peer
        specifierObj[specifierContext] = upstreamsEnum[upstreamIdx][0];

        for (var peerIdx = 0; peerIdx < upstream[metric_name].length; peerIdx++) {
            var peer = upstream[metric_name][peerIdx];
            nextObj[peer.server] = peer;
            nextObj[peer.server].specifierObj = specifierObj;
        }
    }
    metric_obj.handler = named_metric_callback;
    return process_obj(nextObj, metric_obj);
}

// labeled_metric_handler allows for object metrics to be printed, in which each
// member in the object represents a different "code" that should be added to the
// specifier associated with the metric
function labeled_metric_handler(obj, metric, context, specifierContext) {
    var buffer = [];
    var objEnum = Object.entries(obj);

    for (var i = 0; i < objEnum.length; i++) {
        var iterKey = objEnum[i][0];
        var iterObj = objEnum[i][1];
        var labelObj = iterObj[metric[0]];
        var specifierObj = {}
        if (iterObj.specifierObj != null) {
            specifierObj = iterObj.specifierObj;
        }
        specifierObj[specifierContext] = iterKey;
        buffer.push(build_labeled_metrics(labelObj, metric, context, specifierObj));
        delete specifierObj[specifierContext];
    }
    return buffer.join("");

}

// build_labeled_metrics builds the metrics and adds the "code" to the specifier
function build_labeled_metrics(labelObj, metric, context, specifierObj) {
    var buffer = [];
    var labelMetrics = metric[1].metrics;

    for (var i = 0; i < labelMetrics.length; i++) {
        var labelMetric = labelMetrics[i];
        specifierObj.code = labelMetric;
        buffer.push(build_metric(context, metric[0], labelObj[labelMetric],
            specifierObj));
    }
    delete specifierObj.code;
    return buffer.join("");
}

// prepend_metric_handler allows for object metrics to be printed, in which each
// member is appended to the base object name and which corresponds to an object
// in the subrequest response
function prepend_metric_handler(obj, metric, context, specifierContext) {
    var buffer = [];
    var objEnum = Object.entries(obj);
    var prependMetrics = metric[1].metrics;
    var prependMetricsEnum = Object.entries(prependMetrics);

    for (var i = 0; i < prependMetricsEnum.length; i++) {
        var prependMetric = prependMetricsEnum[i];
        var printableMetric = {};
        printableMetric[metric[0] + "_" + prependMetric[0]] = prependMetric[1];
        if (prependMetric[1].help != null) {
            buffer.push(print_metric_help_text(context,
                Object.entries(printableMetric)[0]));
        }

        for (var j = 0; j < objEnum.length; j++) {
            var iterKey = objEnum[j][0];
            var iterObj = objEnum[j][1];
            var specifierObj = {};
            if (iterObj.specifierObj != null) {
                specifierObj = iterObj.specifierObj;
            }
            specifierObj[specifierContext] = iterKey;
            var prependMetricValue = iterObj[metric[0]][prependMetric[0]];
            buffer.push(build_metric(context, metric[0] + "_" + prependMetric[0],
                prependMetricValue, specifierObj));
            delete specifierObj[specifierContext];
        }
    }
    return buffer.join("");
}

// convert_state converts the metric "state" value to a prometheus-compliant value
function convert_state(metric) {
    switch (metric) {
        case "up":
            return 1;
        case "draining":
            return 2;
        case "down":
            return 3;
        case "unavail":
            return 4;
        case "checking":
            return 5;
        case "unhealthy":
            return 6;
        default:
            return 0;
    }
}

// build_metric returns the prometheus-formatted metric for a particular NGINX
// metric
function build_metric(context, metric, metric_val, specifierObj) {
    var specifier = "";

    if (specifierObj != null) {
        specifier = build_specifier(specifierObj);
    }
    if (metric_val == undefined) {
        metric_val = 0;
    }
    return NGINX_PLUS_CONTEXT + "_" + context + "_" + metric + " " +
        specifier + metric_val + "\n";
}

// build_specifier returns the JSON-like prometheus specifier associated with a
// metric
function build_specifier(specifierObj) {
    return JSON.stringify(specifierObj).replace(/\"([^(\")"]+)\":/g, "$1=") + " ";
}

// print_metric_help_text returns the "HELP" and "TYPE" text associated with a
// metric
function print_metric_help_text(context, metric) {
    var metric_context = NGINX_PLUS_CONTEXT + "_" + context + "_" + metric[0];
    return "# HELP " + metric_context + " " + metric[1].help + "\n" +
        "# TYPE " + metric_context + " " + metric[1].type + "\n";
}

// ngx_send_response returns the response through the NGINX response object with
// the built response buffer
function ngx_send_response(r) {
    r.return(200, responseBuffer.join(""));
}

export default {metrics};
